/*
 * [FieldType.java]
 *
 * Summary: Enumeration of various data types for which we handle interconversions.
 *
 * Copyright: (c) 2009-2011 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.5+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.0 2009-04-22 initial release
 */
package com.mindprod.comparatorcutter;

import com.mindprod.common15.STA;
import com.mindprod.fastcat.FastCat;

/**
 * Enumeration of various data types for which we handle interconversions.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 1.0 2009-04-22 initial release
 * @since 2009
 */
@SuppressWarnings( { "UnusedDeclaration" } )
public enum FieldType
    {
        // templates defined for ascending
        SPARE( "-- spare --", "", "", "" ),
        CASE_SENSITIVE_STRING( "String (case sensitive)",
                "int", "case sensitively",
                "%first%getter.compareTo( %second%getter )" ),
        CASE_INSENSITIVE_STRING( "String (case insensitive)",
                "int", "case insensitively",
                "%first%getter.compareToIgnoreCase( %second%getter )" ),
        STRING_LENGTH( "String (length)",
                "int", "length",
                "%first%getter.length() - %second%getter.length()" ),
        COMPARABLE( "Comparable Object",
                "int", "Comparable",
                "%first%getter.compareTo( %second%getter )" ),
        BOOLEAN( "boolean",
                "int", "logically",
                "( %first%getter ? 1 : 0 ) - ( %second%getter ? 1 : 0 )" ),
        SIGNEDBYTE( " byte",
                "int", "numerically",
                "%first%getter - %second%getter" ),
        UNSIGNEDBYTE( "/*unsigned*/ byte",
                "int", "numerically",
                "( %first%getter & 0xff ) - ( %second%getter & 0xff )" ),
        SHORT( "short",
                "int", "numerically",
                "%first%getter - %second%getter" ),
        CHAR( "char",
                "int", "alphabetically",
                "%first%getter - %second%getter" ),
        INT( "int",
                "int", "numerically",
                "%first%getter - %second%getter" ),
        LONG( "long",
                "long", "numerically",
                "%first%getter - %second%getter" ),
        FLOAT( "float",
                "float", "numerically",
                "%first%getter - %second%getter" ),
        DOUBLE( "double",
                "double", "numerically",
                "%first%getter - %second%getter" );

    // ------------------------------ FIELDS ------------------------------

    /**
     * template for how to compare this type of field. contains replacement strings %fieldName %first %second
     * Gives result for ascending.
     */
    private final String comparatorTemplate;

    /**
     * what is Java type for diff variable to hold intermediate compare result?
     */
    private final String diffType;

    /**
     * Java type name
     */
    private final String fieldTypeName;

    /**
     * how you would describe the sorting order. e.g. case sensitively, case insensitively, numerically...
     */
    private final String sortType;

    // -------------------------- PUBLIC INSTANCE  METHODS --------------------------

    /**
     * generate Java code to compare two fields,
     * when it is not the last field being compared.
     *
     * @param useGenerics       true if want code generated using generics
     * @param isComparator      true if generated a Comparable, false if a Comparator
     * @param oClass            the class of the objects being compared.
     * @param isFieldDescending true if want a descending comparison
     * @param fieldName         name of field being compared.
     * @param unique            number to generate unique variables 1..n
     *
     * @return Java source code
     */
    @SuppressWarnings( { "SameParameterValue" } )
    public String generateIntermediateCompareToLine( final boolean useGenerics, boolean isComparator, final String oClass, boolean isFieldDescending, String fieldName, final int unique )
        {
        final FastCat sb = new FastCat( 17 );
        indent( sb, 2 );
        sb.append( diffType );
        sb.append( " " );
        final String diffVar;
        if ( unique == 0 )
            {
            diffVar = "diff";
            }
        else
            {
            diffVar = "diff" + Integer.toString( unique );
            }
        sb.append( diffVar );
        sb.append( " = " );
        sb.append( expandCompareToTemplate( useGenerics, isComparator, oClass, isFieldDescending, fieldName )
        );
        sb.append( ";\n" );
        indent( sb, 2 );
        sb.append( "if ( " );
        sb.append( diffVar );
        sb.append( " != 0 )\n" );
        indent( sb, 3 );
        sb.append( "{\n" );
        indent( sb, 3 );
        sb.append( generateCompareToLine( useGenerics, diffVar ) );
        indent( sb, 3 );
        sb.append( "}\n" );
        return sb.toString();
        }

    /**
     * generate Java code to compare two fields, when it is the last field being compared.
     *
     * @param useGenerics       true if want code generated using generics
     * @param isComparable      true if want a code for a Comparable, false for a Comparator
     * @param oClass            the class of the objects being compared.
     * @param isFieldDescending true if want a descending comparison
     * @param fieldName         name of sort key field
     *
     * @return Code customised to compare two fields of this fieldType
     */
    public String generateLastCompareToLine( final boolean useGenerics, boolean isComparable, final String oClass, boolean isFieldDescending, String fieldName )
        {
        final FastCat sb = new FastCat( 2 );
        final String diffExp = expandCompareToTemplate( useGenerics, isComparable, oClass, isFieldDescending, fieldName );
        indent( sb, 2 );
        sb.append( generateCompareToLine( useGenerics, diffExp ) );
        return sb.toString();
        }

    /**
     * get the Java type of the diff variable
     *
     * @return "int", "long" "float" "double"
     */
    public String getDiffType()
        {
        return diffType;
        }

    /**
     * how you would describe the sorting order.
     *
     * @return e.g. case sensitively, case insensitively, numerically
     */
    public String getSortType()
        {
        return sortType;
        }

    /**
     * string for JComboBox display
     *
     * @return field type name
     */
    public String toString()
        {
        return fieldTypeName;
        }

    // -------------------------- STATIC METHODS --------------------------

    /**
     * indent generated text
     *
     * @param sb    StringBuilder where we are accumulating generated text.
     * @param level How many levels of indentation, in groups of 4 cols.
     */
    private static void indent( FastCat sb, int level )
        {
        sb.append( STA.spaces( level * 4 ) );
        }

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * constructor
     *
     * @param fieldTypeName      Java type name
     * @param diffType           Java type of diff variable used to hold result of compare, or raw return type before signum.
     * @param sortType           how you would describe the sorting order. e.g. case sensitively, case insensitively, numerically
     * @param comparatorTemplate template code to generate a comparison
     */
    FieldType( String fieldTypeName, String diffType, String sortType, String comparatorTemplate )
        {
        this.fieldTypeName = fieldTypeName;
        this.diffType = diffType;
        this.sortType = sortType;
        this.comparatorTemplate = comparatorTemplate;
        }

    // -------------------------- OTHER METHODS --------------------------

    /**
     * replace variables in comparison template
     *
     * @param useGenerics       true if want code generated using generics
     * @param isComparable      true if want a code for a Comparable, false for a Comparator
     * @param oClass            the class of the objects being compared.
     * @param isFieldDescending true if want a descending comparison
     * @param fieldName         name of sort key field
     *
     * @return Code customised to compare two fields of this fieldType
     */
    private String expandCompareToTemplate( final boolean useGenerics, final boolean isComparable, final String oClass, final boolean isFieldDescending, final String fieldName )
        {
        final String first;
        final String second;
        if ( useGenerics )
            {
            if ( isComparable )
                {
                if ( isFieldDescending )
                    {
                    first = "other";
                    second = "this";
                    }
                else
                    {
                    first = "this";
                    second = "other";
                    }
                }
            else
                {
                // comparator
                if ( isFieldDescending )
                    {
                    first = "b";
                    second = "a";
                    }
                else
                    {
                    first = "a";
                    second = "b";
                    }
                }
            }
        else
            {
            // no generics
            if ( isComparable )
                {
                if ( isFieldDescending )
                    {
                    first = "( ( " + oClass + " ) other )";
                    second = "this";
                    }
                else
                    {
                    first = "this";
                    second = "( ( " + oClass + " ) other )";
                    }
                }
            else
                {
                // comparator
                if ( isFieldDescending )
                    {
                    first = "( ( " + oClass + " ) b )";
                    second = "( ( " + oClass + " ) a )";
                    }
                else
                    {
                    first = "( ( " + oClass + " ) a )";
                    second = "( ( " + oClass + " ) b )";
                    }
                }
            }
        final String getter = ( fieldName.length() == 0 ) ? "" : "." + fieldName;
        return comparatorTemplate.replace( "%first", first ).replace( "%second", second ).replace( "%getter", getter );
        }

    /**
     * Compare one field for the Comparable/Comparator
     *
     * @param useGenerics true if want code generated using generics
     * @param diffExp     e.g. diff or Misc.signum( diff ) or Long.signum( diff )
     *
     * @return code to compare fields
     */
    private String generateCompareToLine( final boolean useGenerics, final String diffExp )
        {
        final StringBuilder sb = new StringBuilder( 100 );
        if ( diffType.equals( "int" ) )
            {
            //  "return a.x - b.x;"  or "return this.x - other.x"
            sb.append( "return " );
            sb.append( diffExp );
            sb.append( ";\n" );
            }
        else
            {
            // e.g. "return Long.signum( a.x - b.x );" or "return Long.signum( this.x - other.x );"
            sb.append( "return " );
            if ( useGenerics && diffType.equals( "long" ) )
                {
                sb.append( "Long" );
                }
            else
                {
                sb.append( "Misc" );
                }
            sb.append( ".signum( " );
            sb.append( diffExp );
            sb.append( " );\n" );
            }
        return sb.toString();
        }
    }
